@@ -222,6 +222,11 @@ build_up_reference (type, arg, flags)
 				(flags & LOOKUP_PROTECTED_OK) ? 3 : 2);
       if ((flags & LOOKUP_PROTECT) && basetype == error_mark_node)
 	return error_mark_node;
+      if (basetype == NULL_TREE)
+	{
+	  error_not_base_type (target_type, argtype);
+	  return error_mark_node;
+	}
     }
 
   switch (TREE_CODE (arg))
@@ -456,6 +461,22 @@ convert_to_reference (decl, reftype, expr, strict, flags)
 	    error ("ambiguous pointer conversion");
 	  return rval;
 	}
+      else if (rval = build_type_conversion (CONVERT_EXPR, type, expr, 1))
+	{
+	  if (TYPE_NEEDS_DESTRUCTOR (type))
+	    rval = cleanup_after_call (rval);
+	  else
+	    {
+	      decl = get_temp_name (type, 0);
+	      rval = build (INIT_EXPR, type, decl, rval);
+	      rval = build (COMPOUND_EXPR, reftype, rval,
+			    convert_to_reference (NULL_TREE, reftype, decl,
+						  strict, flags));
+
+	    }
+	  return rval;
+	}
+
       if (form == REFERENCE_TYPE
 	  && type != intype
 	  && TYPE_LANG_SPECIFIC (intype)
@@ -474,14 +495,15 @@ convert_to_reference (decl, reftype, expr, strict, flags)
 
      @@ The second disjunct is provided to make references behave
      @@ as some people think they should, i.e., an interconvertability
-     @@ between references to builtin types.  I am beginning
-     @@ to think that this is wrong.  */
+     @@ between references to builtin types (such as short and
+     @@ unsigned short).  There should be no conversion between
+     @@ types whose codes are different, or whose sizes are different.  */
 
   if (((IS_AGGR_TYPE (type) || IS_AGGR_TYPE (intype))
        && comptypes (type, intype, strict))
       || (!IS_AGGR_TYPE (type)
-	  && !IS_AGGR_TYPE (intype)
-	  && int_size_in_bytes (type) <= int_size_in_bytes (intype)))
+	  && TREE_CODE (type) == TREE_CODE (intype)
+	  && int_size_in_bytes (type) == int_size_in_bytes (intype)))
     {
       /* If EXPR is of aggregate type, and is really a CALL_EXPR,
 	 then we don't need to convert it to reference type if
@@ -569,14 +591,6 @@ convert_from_reference (val)
 	  return val;
 	}
 
-      if (TREE_READONLY (type))
-	{
-	  val = build_indirect_ref (val, "reference bashing (compiler error)");
-	  if (TREE_CODE (val) == VAR_DECL && TREE_READONLY (val)
-	      && DECL_INITIAL (val))
-	    return DECL_INITIAL (val);
-	}
-
       val = build1 (INDIRECT_REF, TYPE_MAIN_VARIANT (dt), val);
 
       TREE_THIS_VOLATILE (val) = TREE_VOLATILE (dt);
@@ -609,10 +623,11 @@ convert_to_real (type, expr)
     {
       tree rval;
       rval = build_type_conversion (CONVERT_EXPR, type, expr, 1);
-      if (rval) return rval;
+      if (rval)
+	return rval;
+      else
+	error ("aggregate value used where a floating point value was expected");
     }
-  else
-    error ("aggregate value used where a floating point value was expected");
 
   {
     register tree tem = make_node (REAL_CST);
@@ -1144,6 +1159,30 @@ convert_pointer_to (type, expr)
   TREE_LITERAL (rval) = TREE_LITERAL (expr);
   return rval;
 }
+
+/* Same as above, but don't abort if we get an "ambiguous" baseclass.
+   There's only one virtual baseclass we are looking for, and once
+   we find one such virtual baseclass, we have found them all.  */
+
+tree
+convert_pointer_to_vbase (type, expr)
+     tree type;
+     tree expr;
+{
+  tree intype = TREE_TYPE (TREE_TYPE (expr));
+  int i;
+
+  for (i = CLASSTYPE_N_BASECLASSES (intype); i > 0; i--)
+    {
+      tree basetype = CLASSTYPE_BASECLASS (intype, i);
+      if (type == basetype)
+	return convert_pointer_to (type, expr);
+      if (value_member (TYPE_MAIN_VARIANT (type),
+			CLASSTYPE_VBASECLASSES (basetype)))
+	return convert_pointer_to_vbase (type, convert_pointer_to (TYPE_MAIN_VARIANT (basetype), expr));
+    }
+  abort ();
+}
 
 /* Create an expression whose value is that of EXPR,
    converted to type TYPE.  The TREE_TYPE of the value
@@ -1378,11 +1417,15 @@ build_type_conversion (code, xtype, expr, for_sure)
   tree type, type_default;
   tree typename = build_typename_overload (xtype), *typenames;
   int n_variants = 0;
-  tree basetype = TREE_TYPE (expr), save_basetype;
+  tree basetype, save_basetype;
   tree rval;
   int exact_conversion = for_sure >= 2;
   for_sure &= 1;
 
+  if (expr == error_mark_node)
+    return error_mark_node;
+
+  basetype = TREE_TYPE (expr);
   if (TREE_CODE (basetype) == REFERENCE_TYPE)
     basetype = TREE_TYPE (basetype);
 
@@ -1478,7 +1521,7 @@ build_type_conversion (code, xtype, expr, for_sure)
     return NULL_TREE;
 
   /* No perfect match found, try default.  */
-  if (TREE_CODE (type) == POINTER_TYPE)
+  if (code == CONVERT_EXPR && TREE_CODE (type) == POINTER_TYPE)
     type_default = ptr_type_node;
   else if (type == void_type_node)
     return NULL_TREE;
@@ -1491,11 +1534,12 @@ build_type_conversion (code, xtype, expr, for_sure)
       type_default = TREE_TYPE (tmp);
     }
 
+  basetype = save_basetype;
+
   if (type_default != type)
     {
       type = type_default;
       typename = build_typename_overload (type);
-      basetype = save_basetype;
 
       while (TYPE_HAS_CONVERSION (basetype))
 	{
@@ -1506,19 +1550,22 @@ build_type_conversion (code, xtype, expr, for_sure)
 	  else break;
 	}
     }
-  else if (type == ptr_type_node)
+
+ try_pointer:
+
+  if (type == ptr_type_node)
     {
       /* Try converting to some other pointer type
-	 with which void* is compatible.  */
-      basetype = save_basetype;
+	 with which void* is compatible, or in situations
+	 in which void* is appropriate (such as &&,||, and !).  */
 
       while (TYPE_HAS_CONVERSION (basetype))
 	{
-	  if (CLASSTYPE_PTR_CONVERSION (basetype) != 0)
+	  if (CLASSTYPE_CONVERSION (basetype, ptr_conv) != 0)
 	    {
-	      if (CLASSTYPE_PTR_CONVERSION (basetype) == error_mark_node)
+	      if (CLASSTYPE_CONVERSION (basetype, ptr_conv) == error_mark_node)
 		return error_mark_node;
-	      typename = DECL_ORIGINAL_NAME (CLASSTYPE_PTR_CONVERSION (basetype));
+	      typename = DECL_ORIGINAL_NAME (CLASSTYPE_CONVERSION (basetype, ptr_conv));
 	      return build_type_conversion_1 (xtype, basetype, expr, typename, for_sure);
 	    }
 	  if (CLASSTYPE_N_BASECLASSES (basetype))
@@ -1526,21 +1573,20 @@ build_type_conversion (code, xtype, expr, for_sure)
 	  else break;
 	}
     }
-  else if (TREE_CODE (type) == POINTER_TYPE
-	   && TREE_READONLY (TREE_TYPE (type))
-	   && TYPE_MAIN_VARIANT (TREE_TYPE (type)) == void_type_node)
+  if (TREE_CODE (type) == POINTER_TYPE
+      && TREE_READONLY (TREE_TYPE (type))
+      && TYPE_MAIN_VARIANT (TREE_TYPE (type)) == void_type_node)
     {
       /* Try converting to some other pointer type
 	 with which const void* is compatible.  */
-      basetype = save_basetype;
 
       while (TYPE_HAS_CONVERSION (basetype))
 	{
-	  if (CLASSTYPE_CONSTPTR_CONVERSION (basetype) != 0)
+	  if (CLASSTYPE_CONVERSION (basetype, constptr_conv) != 0)
 	    {
-	      if (CLASSTYPE_CONSTPTR_CONVERSION (basetype) == error_mark_node)
+	      if (CLASSTYPE_CONVERSION (basetype, constptr_conv) == error_mark_node)
 		return error_mark_node;
-	      typename = DECL_ORIGINAL_NAME (CLASSTYPE_CONSTPTR_CONVERSION (basetype));
+	      typename = DECL_ORIGINAL_NAME (CLASSTYPE_CONVERSION (basetype, constptr_conv));
 	      return build_type_conversion_1 (xtype, basetype, expr, typename, for_sure);
 	    }
 	  if (CLASSTYPE_N_BASECLASSES (basetype))
@@ -1548,7 +1594,34 @@ build_type_conversion (code, xtype, expr, for_sure)
 	  else break;
 	}
     }
+  /* Use the longer or shorter conversion that is appropriate.  */
+  if (TREE_CODE (type) == INTEGER_TYPE
+      && TYPE_HAS_INT_CONVERSION (basetype)
+      && CLASSTYPE_CONVERSION (basetype, int_conv) != error_mark_node)
+    {
+      typename = DECL_ORIGINAL_NAME (CLASSTYPE_CONVERSION (basetype, int_conv));
+      return build_type_conversion_1 (xtype, basetype, expr, typename, for_sure);
+    }
+  if (TREE_CODE (type) == REAL_TYPE
+      && TYPE_HAS_REAL_CONVERSION (basetype)
+      && CLASSTYPE_CONVERSION (basetype, real_conv) != error_mark_node)
+    {
+      typename = DECL_ORIGINAL_NAME (CLASSTYPE_CONVERSION (basetype, real_conv));
+      return build_type_conversion_1 (xtype, basetype, expr, typename, for_sure);
+    }
 
+  /* THIS IS A KLUDGE.  */
+  if (TREE_CODE (type) != POINTER_TYPE
+      && (code == TRUTH_ANDIF_EXPR
+	  || code == TRUTH_ORIF_EXPR
+	  || code == TRUTH_NOT_EXPR))
+    {
+      /* Here's when we can convert to a pointer.  */
+      type = ptr_type_node;
+      goto try_pointer;
+    }
+
+  /* THESE ARE TOTAL KLUDGES.  */
   /* Default promotion yields no new alternatives, try
      conversions which are anti-default, such as
 
@@ -1604,10 +1677,6 @@ build_type_conversion (code, xtype, expr, for_sure)
     else if (TREE_CODE (type) == INTEGER_TYPE)
       if (TYPE_HAS_REAL_CONVERSION (basetype))
 	type = double_type_node;
-      else if (code == TRUTH_ANDIF_EXPR
-	       || code == TRUTH_ORIF_EXPR
-	       || code == TRUTH_NOT_EXPR)
-	type = ptr_type_node;
       else
 	return NULL_TREE;
     else
@@ -1645,6 +1714,11 @@ build_default_binary_type_conversion (code, arg1, arg2)
   tree type2 = TREE_TYPE (*arg2);
   char *name1, *name2;
 
+  if (TREE_CODE (type1) == REFERENCE_TYPE)
+    type1 = TREE_TYPE (type1);
+  if (TREE_CODE (type2) == REFERENCE_TYPE)
+    type2 = TREE_TYPE (type2);
+
   if (TREE_CODE (TYPE_NAME (type1)) != TYPE_DECL)
     {
       tree decl = typedecl_for_tag (type1);
@@ -1672,7 +1746,7 @@ build_default_binary_type_conversion (code, arg1, arg2)
   if (! TYPE_HAS_CONVERSION (type1))
     {
       if (! TYPE_HAS_CONVERSION (type2))
-	error ("type conversion required for types `%s' and `%s'",
+	error ("type conversion required for binary operation on types `%s' and `%s'",
 	       name1, name2);
       else
 	error ("type conversion required for type `%s'", name1);
