@@ -357,14 +357,14 @@ layout_vbasetypes (rec, max)
   return max;
 }
 
+#if 0
+/* This function should never be needed.  */
 void
 fixup_vbase_offsets (type)
      tree type;
 {
   tree virtuals = TREE_CHAIN (CLASS_ASSOC_VIRTUALS (type));
 
-  return;
-
   while (virtuals)
     {
       tree pfn = FNADDR_FROM_VTABLE_ENTRY (TREE_VALUE (virtuals));
@@ -393,6 +393,7 @@ fixup_vbase_offsets (type)
       virtuals = TREE_CHAIN (virtuals);
     }
 }
+#endif
 
 /* Lay out the base types of a record type, REC.
    Tentatively set the size and alignment of REC
@@ -435,6 +436,7 @@ layout_basetypes (rec)
       if (TYPE_SIZE (basetype) == 0)
 	{
 	  error_with_aggr_type (basetype, "base class `%s' has incomplete type");
+	  SET_CLASSTYPE_VIAS (rec, i, 1, 0);
 	  continue;
 	}
 
@@ -472,6 +474,8 @@ layout_basetypes (rec)
 	  decl = build_lang_decl (FIELD_DECL, get_identifier (name),
 				  build_pointer_type (basetype));
 	  DECL_FIELD_CONTEXT (decl) = rec;
+	  SET_DECL_FCONTEXT (decl, TYPE_MAIN_VARIANT (basetype));
+	  DECL_VBASE_P (decl) = 1;
 	  TREE_CHAIN (decl) = vbase_decls;
 	  vbase_decls = decl;
 
@@ -683,6 +687,7 @@ hash_tree_cons (via_public, via_virtual, purpose, value, chain)
   return t;
 }
 
+/* Constructor for hashed lists.  */
 tree
 hash_tree_chain (value, chain)
      tree value, chain;
@@ -699,6 +704,56 @@ hash_tree_chain (value, chain)
   return t;
 }
 
+/* Similar, but used for concatenating two lists.  */
+tree
+hash_chainon (list1, list2)
+     tree list1, list2;
+{
+  if (list2 == 0)
+    return list1;
+  if (list1 == 0)
+    return list2;
+  if (TREE_CHAIN (list1) == NULL_TREE)
+    return hash_tree_chain (TREE_VALUE (list1), list2);
+  return hash_tree_chain (TREE_VALUE (list1),
+			  hash_chainon (TREE_CHAIN (list1), list2));
+}
+
+tree
+build_decl_list_1 (value)
+     tree value;
+{
+  tree list = NULL_TREE;
+
+  if (TREE_CODE (value) == IDENTIFIER_NODE)
+    {
+      list = IDENTIFIER_AS_LIST (value);
+      if (list != NULL_TREE
+	  && (TREE_CODE (list) != TREE_LIST
+	      || TREE_VALUE (list) != value))
+	list = NULL_TREE;
+      else if (TREE_TYPE (value) != NULL_TREE
+	       && TREE_CODE (TREE_TYPE (TREE_TYPE (value))) == RECORD_TYPE)
+	{
+	  tree type = TREE_TYPE (TREE_TYPE (value));
+	  if (CLASSTYPE_AS_ID_LIST (type) == NULL_TREE)
+	    CLASSTYPE_AS_ID_LIST (type) = perm_tree_cons (NULL_TREE, value, NULL_TREE);
+	  list = CLASSTYPE_AS_ID_LIST (type);
+	}
+    }
+  else if (TREE_CODE (value) == RECORD_TYPE
+	   && TYPE_LANG_SPECIFIC (value))
+    list = CLASSTYPE_AS_LIST (value);
+
+  if (list != NULL_TREE)
+    {
+      assert (TREE_CHAIN (list) == NULL_TREE);
+      return list;
+    }
+
+  return build_decl_list (NULL_TREE, value);
+}
+
 /* Look in the type hash table for a type isomorphic to
    `build_tree_list (NULL_TREE, VALUE)'.
    If one is found, return it.  Otherwise return 0.  */
@@ -710,7 +765,37 @@ list_hash_lookup_or_cons (value)
   register int hashcode = TYPE_HASH (value);
   register struct list_hash *h;
   struct obstack *ambient_obstack;
-  tree list;
+  tree list = NULL_TREE;
+
+  if (TREE_CODE (value) == IDENTIFIER_NODE)
+    {
+      list = IDENTIFIER_AS_LIST (value);
+      if (list != NULL_TREE
+	  && (TREE_CODE (list) != TREE_LIST
+	      || TREE_VALUE (list) != value))
+	list = NULL_TREE;
+      else if (TREE_TYPE (value) != NULL_TREE
+	       && TREE_CODE (TREE_TYPE (TREE_TYPE (value))) == RECORD_TYPE)
+	{
+	  tree type = TREE_TYPE (TREE_TYPE (value));
+	  if (CLASSTYPE_AS_ID_LIST (type) == NULL_TREE)
+	    CLASSTYPE_AS_ID_LIST (type) = perm_tree_cons (NULL_TREE, value, NULL_TREE);
+	  list = CLASSTYPE_AS_ID_LIST (type);
+	}
+    }
+  else if (TREE_CODE (value) == TYPE_DECL
+	   && TREE_CODE (TREE_TYPE (value)) == RECORD_TYPE
+	   && TYPE_LANG_SPECIFIC (TREE_TYPE (value)))
+    list = CLASSTYPE_AS_ID_LIST (TREE_TYPE (value));
+  else if (TREE_CODE (value) == RECORD_TYPE
+	   && TYPE_LANG_SPECIFIC (value))
+    list = CLASSTYPE_AS_LIST (value);
+
+  if (list != NULL_TREE)
+    {
+      assert (TREE_CHAIN (list) == NULL_TREE);
+      return list;
+    }
 
   if (debug_no_list_hash)
     return hash_tree_chain (value, NULL_TREE);
@@ -819,6 +904,35 @@ assoc_value (elem, type)
   return rval;
 }
 
+tree
+virtual_member (elem, list)
+     tree elem;
+     tree list;
+{
+  tree t;
+  tree rval, nval;
+
+  for (t = list; t; t = TREE_CHAIN (t))
+    if (elem == TREE_VALUE (t))
+      return t;
+  rval = 0;
+  for (t = list; t; t = TREE_CHAIN (t))
+    {
+      int i;
+      for (i = CLASSTYPE_N_BASECLASSES (TREE_TYPE (t)); i > 0; i--)
+	{
+	  nval = assoc_value (elem, CLASSTYPE_BASECLASS (TREE_TYPE (t), i));
+	  if (nval)
+	    {
+	      if (rval && TREE_TYPE (nval) != TREE_TYPE (rval))
+		abort ();
+	      rval = nval;
+	    }
+	}
+    }
+  return rval;
+}
+
 void
 debug_dump_assoc (elem)
      tree elem;
@@ -851,7 +965,6 @@ debug_dump_assoc (elem)
     }
 }
 
-
 char *
 lang_printable_name (decl)
      tree decl;
@@ -860,7 +973,11 @@ lang_printable_name (decl)
       || DECL_LANG_SPECIFIC (decl) == 0)
     {
       if (DECL_NAME (decl))
-	return IDENTIFIER_POINTER (DECL_NAME (decl));
+	{
+	  if (THIS_NAME_P (DECL_NAME (decl)))
+	    return "this";
+	  return IDENTIFIER_POINTER (DECL_NAME (decl));
+	}
       return "((anonymous))";
     }
   if (DECL_PRINT_NAME (decl) == 0)
@@ -912,6 +1029,9 @@ lang_output_debug_info (type)
 	 which does not have its implementation defined here.  */
       if (TYPE_VIRTUAL_P (type) && write_virtuals == 2)
 	return value_member (DECL_NAME (TYPE_NAME (type)), pending_vtables) != 0;
+      if (CLASSTYPE_INTERFACE_ONLY (type))
+	return 0;
+
       return CLASSTYPE_ASM_WRITTEN (type);
     }
   else
